# 基于FDU-OS的实时调度实现 (模块实验创新)

**分支介绍**
- MF-lazy(主分支)
- RMS-lss(未使用懒标签，可进行CR的测试)
- Base(课程架构的基础，可进行课程lab的测试)

在课程框架(框架以及完成的任务写在后面)的基础上，尝试实现了RTOS的部分内容，主要有如下内容：

- proc结构体中加入优先级，周期，运行时间等与实时相关的属性
- 实现RMS调度，选择周期最短的进程
- 实现临界资源结构体，并使用Ceiling解决优先级反转的问题(RMS-lss分支中可测试)
- 实现了多重队列，引入懒标签机制，优化访问次数。

为了简化系统，我们采用了多核独立的方式实现，即每个CPU核独自享有一个调度队列，他们之间只有CR是共享的。

## proc 结构体
位于./src/kernel/proc.h中


## timer_handler
每次进程中断，都需要更新所有进程的当前时间current_time，以及当前进程的已运行时间
## RMS调度

在进程初始化时，传入需要运行时间(exec_time)以及周期(period)进行初始化，模拟RTOS中预测运行时间的行为。

以MAX(20-period/100,0)作为优先级，分配进入对应的队列，多重队列总共25级，其中0到20为实时部分，-4到-1为非实时的(本项目中暂未使用，未来可整合在一起)

调度时，从优先级为20的队列开始寻找，如果不为空，则进入后遍历，选取周期最短的进程。

由于课程架构中每个进程享有固定长度的时间片，且时间片耗尽后进行下一次调度，所以在这种情况下，RMS会出现抢占的情况，即上一个低优先级任务时间片耗尽了，但没完成，又到了高优先级任务的新周期，这时候高优先级就会抢占低优先级的cpu资源。

运行效果：

![RMS-preemption](./images/preemption.png)

可以看到高优先级的进程一旦准备就绪，会立刻进行抢占。(这是使用懒标签后的演示图)

## 临界资源

```c
typedef struct CR {
    SleepLock lock;        // 睡眠锁
    ListNode node;
    int priority;           // 资源的优先级
    int cid;              // 资源的ID
} CR;
```

在内核初始化20个临界资源,并让子进程尝试获取。基于此实现了Ceiling的算法以解决优先级反转的问题，为了简单起见，所有CR资源的优先级都设置为最高(20)。

运行效果：

![CR](./images/CR.png)

可以看到一个低优先级的任务(14)先获取了CR资源，优先级提升为(20)，这时一个高优先级任务(18)也要获取CR资源，但必须要等待低优先级完成，如果没有Ceiling，就导致了优先级反转，使用Ceiling解决了这个问题。

## 带有懒标签的多重队列

实现位于分支MF-lazy中，引入该方法后还没解决CR部分的错误，所以在MF-lazy中无法进行CR的测试

实现多重队列的动机是减少进程切换时对调度队列的访问次数，但我们发现在RTOS直接使用多重队列其实并没什么用，因为每次切换必须更新current_time，这要求无论如何遍历所有的进程。

但同时我们注意到很多低优先级的任务其实不需要频繁的更新当前时间，只需要在被使用时更新，得知当前时间以及判断是否超时即可。鉴于我们尝试的时软实时操作系统，假设任务miss不会有什么严重后果，所以这个想法应该是可行的。

因此为调度队列引入了一个懒标签Lazy，以及一个完成的调度队列complete，并类似的引入Lazy和Ready标签，后者记录了完成队列中进程最短还有多久就绪。

算法是这样的，在time-handler中进行更新时，依旧从高向低遍历多重队列，如果不为空，则进入该级队列，遍历当中的所有进程，假设级别为level，则更新公式为：
$$
current = (current +elapse+rqLazy[level])\%period
$$
​	其中current为进程当前时间，elapse为时间片长度，rqLazy是每一级的懒标签,period是该进程的周期。

​	更新完这一级后，向下更新懒标签：
$$
rqLazy[level-1]=\left \{
\begin{aligned}
&rqLazy[level-1]+rqLazy[level]+elapse& level>0\\
& 0&else
\end{aligned}
\right.
$$
在此之前，还需要检查comReady和comLazy，如果前者小于等于后者，则需要遍历complete，使用Lazy进行更新，与rq是类似的。

### 效果

这是没有使用lazy前的，多重队列每次handler的平均访问次数：

![withoutLazy](./images/withoutLazy.png)

使用lazy后：

![Lazy](./images/MQwithLzy.png)


访问次数是原始值的**29.73%**，效果显著，且没有破坏调度的逻辑。

## 运行方法

输入make qemu即可

测试文件在./src/test/rbtree_test.c

最后，由于时间所限，没有将实时的部分与课程中的lab整合在一起，所以在MF-lazy和RMS-lss分支中应该是无法进行课程lab中的测试。

以下是框架基础

# OS-24Fall-FDU
Lab repository for OS Fall 2024 at Fudan University

这是复旦大学 2024 年秋季学期《操作系统（H）》课程的配套实验内容。我们将建立一个基于 ARM 架构的简易教学操作系统。

[实验文档](https://osh.fducslg.com)

暂定的实验内容将包括：

* 引导 Booting
* 内存管理
* 进程管理
* 多核管理
* 中断与异常、缺页处理
* 块设备驱动
* 文件系统
* IPC
* Shell

Reference:

- [去年的实验](https://github.com/FDUCSLG/OS-23Fall-FDU/)

- [Arm® Architecture Reference Manual](https://cs140e.sergio.bz/docs/ARMv8-Reference-Manual.pdf)
- [Arm® Instruction Set Reference Guide](https://ipads.se.sjtu.edu.cn/courses/os/reference/arm_isa.pdf)
- [ARM Cortex-A Series Programmer’s Guide for ARMv8-A](https://cs140e.sergio.bz/docs/ARMv8-A-Programmer-Guide.pdf)
- [ARM GCC Inline Assembler Cookbook](https://www.ic.unicamp.br/~celio/mc404-s2-2015/docs/ARM-GCC-Inline-Assembler-Cookbook.pdf)
- [ucore step-by-step](https://1790865014.gitbook.io/ucore-step-by-step/)
- [ucore实验指导书](https://chyyuu.gitbooks.io/ucore_os_docs/content/)


# 各部分实现方法：
    基于学校课程的架构，完成了以下内容（独立编写的部分会在文档中说明）：
## 内存管理
位于./src/mem.c中

实现了一个基于链表和块划分的简单的物理内存分配器。使用Slab分块管理内存,实现函数包括
- kinit
- kalloc_page、kfree_page
- get_allocator_index
- kalloc
- kfree

### 运行结果

评测指标为使用物理页数，以下为连续三次测试的结果。

![lab1-result](./images/lab1-result.png)


## 内核进程

- ./src/aarch64/switch.S
- ./src/aarch64/trap.S
- ./src/kernel/proc.c
- ./src/kernel/sche.c中

实现了如下图所示的进程状态转换逻辑的内核进程部分。

![lab2-trans](./images/lab2-trans.png)

调度算法为FIFO。

### 运行结果

主要测试是否能正确exit和wait：

![lab2-result](./images/lab2-result.png)

## 用户进程

该部分同时完成了页表的相关操作,通过中断进入用户态。

- ./src/kernel/pt.c
- ./src/aarch64/trap.S
- ./src/kernel/proc.c 中新增kill

调度算法是绝对公平的调度,选取等待时间最长的进程（即最先进入队列的）

### 运行结果

测试了页表的功能，以及用户进程

![lab3-vmresult](./images/vmtest.png)
![lab3-user](./images/user_proctest.png)

以上部分测试，可在./src/kernel/core.c,解注释掉kernel_entry的一系列test后，输入make qemu进行测试。

kalloc_test似乎随着课程框架的更新，没有加入编译，无法正确定位。

## 文件系统

只提供接口，使用mock方法进行测试

- ./src/fs/cache.c
- ./src/fs/inode.c

### 运行结果

![lab5-result](./images/OS_lab5.png)
![lab6-result](./images/lab6-result.png)

# 日志

## RMS新增 12.8

### 测试

测试写在rbtree_test.c中，由于无法触发时钟中断，只能人为的使用sched进行中断，并进行timer的更新

### proc新增

可在proc.h中看新增的属性

### RT系列的初始化

proc.c中RT开头的便是新增的初始化函数

### 调度队列

改为多个cpu独自享有一个调度队列，对rq和rqlock做了一些修改

### 调度策略

目前使用的是RMS

## 12.12 新增

### CR资源

加了一下CR资源，在想怎么加入到测试中，测试优先级反转的问题，可能需要先实现简单的fork，否则单核上没有等待

### EDF实现

简单修改了一下priority以及Inschedtime_handler中的更新逻辑，支持EDF调度，可抢占了。


### 后续

实现一个简单的多重反馈队列，RT优先级可离散化到0到20之间。完成优先级反转的测试及使用相关算法(Ceiling和继承)

## RMS新增 12.15

### fork和vm的deep_copy

由于要测试优先级，所以需要在多核上运行，由于调度队列是单独的，所以尝试使用fork，写了下发现不行，页表错误，感觉debug要很久随放弃，采用了更简单的实现方法：

简单修改一下start_proc函数，让其允许传入cpu id参数以决定在哪个cpu上启动(


fork在proc.c中，deep_copy_vm在pt.c中

## 12.16 新增

### 简单的CR实现以及Ceiling实现


## 12.18 新增

多重反馈队列和RMS(因为EDF的优先级是动态的，使用多重队列毫无意义)

后续再做的话就是优化一下队列吧，因为现在还是需要每个时间片去遍历所有进程去更新时间，可以引入一个lazy标签来优化。