#include <aarch64/intrinsic.h>
#include <kernel/cpu.h>
#include <kernel/printk.h>
#include <kernel/sched.h>
#include <test/test.h>
#include <common/buf.h>
#include <driver/virtio.h>

volatile bool panic_flag;

NO_RETURN void idle_entry()
{
    set_cpu_on();
    while (1) {
        yield();
        if (panic_flag)
            break;
        arch_with_trap
        {
            arch_wfi();
        }
    }
    set_cpu_off();
    arch_stop_cpu();
}

NO_RETURN void kernel_entry()
{
    printk("Hello world! (Core %lld)\n", cpuid());
    //kalloc_test();
    //proc_test();
    // vm_test();
    //user_proc_test();
    // io_test();
    rbtree_test();
    /* LAB 4 TODO 3 BEGIN */
    Buf mbr_buf;
    mbr_buf.block_no = 0;        // 设置读取MBR的块号
    mbr_buf.flags = B_VALID;     // 标记为有效，不需要写入

    // 调用 virtio_blk_rw 读取MBR数据
    if (virtio_blk_rw(&mbr_buf) < 0) {
        printk("Failed to read MBR\n");
        while (1) yield();
    }

    // 等待 I/O 操作完成
    //wait_sem(&mbr_buf.sem);

    // 检查读取是否成功
    if (!(mbr_buf.flags & B_VALID)) {
        printk("Failed to validate MBR\n");
        while (1) yield();
    }

    // 解析MBR，找到第二分区
    //u32 *partition_table = (u32 *)(mbr_buf.data + 446);
    u32 second_partition_lba = *(u32 *)(mbr_buf.data + 446 + 16 + 8);   // 第二分区起始 LBA
    u32 second_partition_size = *(u32 *)(mbr_buf.data + 446 + 16 + 12); // 第二分区大小

    printk("Second partition: LBA = %u, Size = %u\n", second_partition_lba, second_partition_size);

    /* LAB 4 TODO 3 END */

    while (1)
        yield();
}

NO_INLINE NO_RETURN void _panic(const char *file, int line)
{
    printk("=====%s:%d PANIC%lld!=====\n", file, line, cpuid());
    panic_flag = true;
    set_cpu_off();
    for (int i = 0; i < NCPU; i++) {
        if (cpus[i].online)
            i--;
    }
    printk("Kernel PANIC invoked at %s:%d. Stopped.\n", file, line);
    arch_stop_cpu();
}