#pragma once

#include <common/defines.h>
#include <common/list.h>
#include <common/sem.h>
#include <common/rbtree.h>
#include <kernel/pt.h>
#include <kernel/cpu.h>


enum procstate { UNUSED, RUNNABLE, RUNNING, SLEEPING, DEEPSLEEPING, ZOMBIE };

typedef struct UserContext {
    u64 spsr,elr,lr,sp;
    u64 x[32];
} UserContext;


typedef struct KernelContext {
    // TODO: customize your context
    u64 lr,x0,x1;
    u64 x[11]; //x19-29
} KernelContext;

// embeded data for procs
struct schinfo {
    // TODO: customize your sched info
    // 
    ListNode rq;//running runnable
};

#define DEBUG 1

typedef struct Proc {
    bool killed;
    bool idle;
    int pid;//可被回收复用
    int exitcode;
    int wait_time;
    int prim_priority; // 任务原始优先级
    int true_priority; // 任务的真实优先级
    int current_time ;//当前时间，用来重启任务
    int period;
    int exec_time;             // 需要的执行时间
    int execed_time;  // 已执行时间
    bool complete;//完成标志
    enum procstate state;
    Semaphore task_sleep;  // 任务信号量，每隔一个周期post一下
    Semaphore childexit;
    SpinLock plock;//进程锁，猛猛加锁，解决并发问题
    ListNode children;
    ListNode ptnode;
    ListNode crnode;
    struct Proc *parent;
    struct schinfo schinfo;
    struct pgdir pgdir;
    void *kstack;
    UserContext *ucontext;
    KernelContext *kcontext;
#if DEBUG
    int currfuncid;
    int currfuncpos;
#endif
} Proc;

typedef struct PidNode {
    int pid;
    struct PidNode *next;
} PidNode;

typedef struct CR {
    SleepLock lock;        // 睡眠锁
    ListNode node;
    int priority;           // 资源的优先级
    int cid;              // 资源的ID
} CR;

// 初始化 CR 结构体的函数
void init_cr_resources();
void init_critical_resource(CR *cr, int priority,int cid);
void Rinit_kproc();
void init_kproc();
void init_proc(Proc *);
void RT_init(Proc*p,int period,int exec_time);
WARN_RESULT Proc *create_proc();
WARN_RESULT Proc *RT_create(int period,int exec_time);
int start_proc(Proc *, void (*entry)(u64), u64 arg);
NO_RETURN void exit(int code);
WARN_RESULT int wait(int *exitcode);
WARN_RESULT int kill(int pid);
void set_parent_to_this(Proc *proc);

bool acquire_CR(u64 i);
Proc* fork();
int start_proc_with_id(Proc *p, void (*entry)(u64), u64 arg,int cid);
void release_CR(CR *cr);