#include <kernel/sched.h>
#include <kernel/proc.h>
#include <kernel/mem.h>
#include <kernel/printk.h>
#include <aarch64/intrinsic.h>
#include <kernel/cpu.h>
#include <common/rbtree.h>
#include <common/spinlock.h>
#include <common/list.h>

extern bool panic_flag;
extern void swtch(KernelContext *new_ctx, KernelContext **old_ctx);
static SpinLock rqlock[NCPU];
static ListNode rq[NCPU][NUM_PRIORITY];
static ListNode complete[NCPU];
//懒标签，其中调度队列的懒标签记录流逝的时间，完成队列里的懒标签记录剩余唤醒时间，为0则进行唤醒(一定是period-curenttime最小的)。
static int rqLazy[NCPU][NUM_PRIORITY];
static int comLazy[NCPU]; //记录经过的时间
static int comReady[NCPU]; // 记录最早被重启的时间
static struct timer sched_timer[NCPU];
static u64 vist_count = 0;
static int trap_times = 0;
static bool once = false;

void init_sched()
{
    // TODO: initialize the scheduler
    // 1. initialize the resources (e.g. locks, semaphores)
    // 2. initialize the scheduler info of each CPU
    for (int i = 0; i < NCPU; i++) {
        init_spinlock(&rqlock[i]);
        //初始化多重反馈队列
        //-4 到 20，正数和0是实时的
        for (int j = 0; j < NUM_PRIORITY; j++) {
            init_list_node(&rq[i][j]);
            rqLazy[i][j] = 0;
        }
        init_list_node(&complete[i]);
        comLazy[i] = 0;
        comReady[i] = __INT_MAX__;
        struct Proc *p = kalloc(sizeof(struct Proc));
        p->idle = 1;
        p->state = RUNNING;
        cpus[i].sched.thisproc = cpus[i].sched.idle = p;
        sched_timer[i].triggered = false;
        sched_timer[i].data = i;
        sched_timer[i].elapse = 10;
        sched_timer[i].handler = &sched_timer_handler;
    }
}

void sched_timer_handler(struct timer *tim)
{
    thisproc()->currfuncid = 11;
    tim->triggered = false;
    //Proc * t = thisproc();
    //acquire_spinlock(&t->plock);
    // t->execed_time  += tim->elapse;
    // if(!t->idle&&t->execed_time>=t->exec_time){
    //     t->execed_time =0;
    //     t->complete = true;
    // }
    // release_spinlock(&t->plock);
    // _for_in_list(p, &rq[cpuid()]) {
    //     //printk("进入了\n");]
    //     if(p==&rq[cpuid()])
    //         break;
    //     auto proc = container_of(p, struct Proc, schinfo.rq);
    //     acquire_spinlock(&proc->plock);
    //     // 更新等待时间并选择等待时间最长的进程
    //     proc->current_time += tim->elapse;
    //     //printk("Task %d current time:%d\n",proc->pid,proc->current_time);
    //     if(proc->current_time >= proc->period){
    //         proc->current_time = 0;
    //         // if(proc->execed_time>=proc->exec_time)
    //         //     printk("Task %d is completed in period,now set current time:%d",proc->pid,proc->current_time);
    //         //判断一下是不是为true
    //         // if(proc->complete==true)
    //         //     printk("Task %d missed\n",proc->pid);
    //         proc->complete = false;
    //     }
    //     release_spinlock(&proc->plock);
    // }
    acquire_sched_lock();
    if (cpuid() == 0 && !once) {
        once = true;
        printk("Task sequence can be solved\n");
    }
    sched(RUNNABLE);
}

void Insched_timer_handler(struct timer *tim)
{
    thisproc()->currfuncid = 8;
    tim->triggered = false;
    Proc *t = thisproc();
    acquire_spinlock(&t->plock);
    t->execed_time += tim->elapse;
    int count = 0;
    bool release = false;
    trap_times += 1;
    usize cid = cpuid();

    if (complete[cid].next != &complete[cid]) {
        comLazy[cid] += tim->elapse;
    }

    if (comReady[cid] <= comLazy[cid]) {
        //有进程需要被唤醒

        int latest = __INT_MAX__;
        auto p = complete[cid].next;

        while (p != &complete[cid]) {
            vist_count += 1;
            auto proc = container_of(p, struct Proc, schinfo.rq);
            auto temp = p->next;
            proc->current_time += comLazy[cid];

            if (proc->current_time >= proc->period) {
                proc->complete = false;
                proc->current_time = proc->current_time % proc->period;
                _detach_from_list(p);
                _insert_into_list(&rq[cid][proc->true_priority], p);

            } else
                latest = MIN(latest, proc->period - proc->current_time);
            p = temp;
        }
        comReady[cid] = latest;
        comLazy[cid] = 0;
    }

    if (!t->idle && t->execed_time >= t->exec_time) {
        t->execed_time = 0;
        t->complete = true;
        _detach_from_list(&t->schinfo.rq);
        //从这里摘下来，并加入完成队列
        t->current_time += tim->elapse;

        comReady[cid] =
                MIN(comReady[cid] - comLazy[cid], t->period - t->current_time);
        auto p = complete[cid].next;
        //插入新的完成的进程，更新一次队列

        if (comLazy[cid] != 0) {
            while (p != &complete[cid]) {
                vist_count += 1;
                auto proc = container_of(p, struct Proc, schinfo.rq);
                auto temp = p->next;
                proc->current_time += comLazy[cid];

                if (proc->current_time >= proc->period) {
                    proc->complete = true;
                    proc->current_time = proc->current_time % proc->period;

                    _detach_from_list(p);

                    _insert_into_list(&rq[cid][proc->true_priority], p);
                }
                p = temp;
            }
            comLazy[cid] = 0;
        }
        if(p!= &complete[cid])
                _insert_into_list(&complete[cid], &t->schinfo.rq);

        // _for_in_list(p,&t->crnode){
        //     if(p==&t->crnode||count>=6)
        //         break;
        //     count++;
        //     auto cr = container_of(p,struct CR, node);
        //     //_detach_from_list(&cr->node);
        //     release=true;
        //     release_CR(cr);
        // }
        ListNode *pt = t->crnode.next; // 获取列表的第一个元素

        // 逐个遍历链表节点
        while (pt != pt->next) {
            // 处理当前节点
            count++;

            auto cr = container_of(pt, struct CR, node); // 获取当前节点的数据
            // 标记 CR 锁为已释放
            release = true;
            release_CR(cr); // 释放 CR 锁
            // 如果达到了最大释放次数，跳出循环
            if (count >= 6) {
                break;
            }
            pt = pt->next; // 移动到下一个节点
        }

        if (release)
            printk(" before release CR: proc->priority: %d\n",
                   t->true_priority);
        t->true_priority = t->prim_priority;
        if (release)
            printk(" after release CR: proc->priority: %d\n\n\n",
                   t->true_priority);
    }
    release_spinlock(&t->plock);
    for (int level = NUM_PRIORITY - 1; level >= 0; level--) {
        //printk("进入了\n");]
        //如果调度队列为空，则直接跳过
        vist_count++;
        if (&rq[cid][level] == rq[cid][level].next) {
            if (level > 0 && rqLazy[cid][level] > 0) {
                rqLazy[cid][level - 1] += rqLazy[cid][level];
            }
            rqLazy[cid][level] = 0;
            continue;
        }
        auto p = rq[cid][level].next;
        while (p != &rq[cid][level]) {
            vist_count++;
            auto proc = container_of(p, struct Proc, schinfo.rq);
            acquire_spinlock(&proc->plock);
            // 更新等待时间并选择等待时间最长的进程
            proc->current_time += (tim->elapse + rqLazy[cid][level]);

            // 更新优先级

            //printk("Task %d current time:%d\n",proc->pid,proc->current_time);
            if (proc->current_time >= proc->period) {
                proc->current_time = proc->current_time % proc->period;
                //if(proc->complete)
                //   printk("Task %d is completed in period,now set current time:%d\n",proc->pid,proc->current_time);
                //判断一下是不是为true
                if (proc->complete == false)
                    printk("Task %d missed\n", proc->pid);
                proc->complete = false;
            }
            //proc->prim_priority = proc->period - proc->current_time; //EDF
            proc->true_priority = proc->prim_priority;
            release_spinlock(&proc->plock);
            p = p->next;
        }
        //能到这里肯定是进入里
        if (level > 0) {
            rqLazy[cid][level - 1] += (tim->elapse + rqLazy[cid][level]);
            //     printk("rqLazy:%d\n",rqLazy[cid][level]);
            // printk("next rqLazy:%d\n\n",rqLazy[cid][level-1]);
        }

        rqLazy[cid][level] = 0;
        break; //说明进入while了，直接break
    }
    //printk("average visit count:%lld\n", vist_count/trap_times);
}

Proc *thisproc()
{
    // TODO: return the current process
    return cpus[cpuid()].sched.thisproc;
}

void init_schinfo(struct schinfo *p)
{
    // TODO: initialize your customized schinfo for every newly-created process
    init_list_node(&p->rq);
}

void acquire_sched_lock()
{
    // TODO: acquire the sched_lock if need
    acquire_spinlock(&rqlock[cpuid()]);
}

void release_sched_lock()
{
    // TODO: release the sched_lock if need
    release_spinlock(&rqlock[cpuid()]);
}

bool is_zombie(Proc *p)
{
    bool r;
    acquire_sched_lock();
    r = p->state == ZOMBIE;
    release_sched_lock();
    return r;
}

bool is_unused(Proc *p)
{
    bool r;
    acquire_sched_lock();
    r = p->state == UNUSED;
    release_sched_lock();
    return r;
}

bool _activate_proc(Proc *p, bool onalert)
{
    // TODO:(Lab5 new)
    // if the proc->state is RUNNING/RUNNABLE, do nothing and return false
    // if the proc->state is SLEEPING/UNUSED, set the process state to RUNNABLE, add it to the sched queue, and return true
    // if the proc->state is DEEPSLEEPING, do nothing if onalert or activate it if else, and return the corresponding value.
    acquire_sched_lock();

    // 如果进程的状态为 RUNNING 或 RUNNABLE，则不执行任何操作，返回 false
    if (p->state == RUNNING || p->state == RUNNABLE || p->state == ZOMBIE ||
        (p->state == DEEPSLEEPING && onalert)) {
        release_sched_lock();
        return false;
    }

    // 如果进程的状态为 SLEEPING 或 UNUSED，将其状态设置为 RUNNABLE，
    // 重置等待时间，将其加入调度队列并返回 true
    if (p->state == SLEEPING || p->state == UNUSED ||
        (p->state == DEEPSLEEPING && !onalert)) {
        p->state = RUNNABLE;
        p->wait_time = 0;
        _insert_into_list(&rq[cpuid()][p->true_priority],
                          &p->schinfo.rq); // 加入调度队列
        release_sched_lock();
        return true;
    }

    // 如果状态不在以上范围内，触发 PANIC
    release_sched_lock();
    PANIC();

    return false; // 作为防御性返回
}

bool _activate_proc_with_id(Proc *p, bool onalert, int cid)
{
    // TODO:(Lab5 new)
    // if the proc->state is RUNNING/RUNNABLE, do nothing and return false
    // if the proc->state is SLEEPING/UNUSED, set the process state to RUNNABLE, add it to the sched queue, and return true
    // if the proc->state is DEEPSLEEPING, do nothing if onalert or activate it if else, and return the corresponding value.
    acquire_sched_lock();
    thisproc()->currfuncid = 7;
    // 如果进程的状态为 RUNNING 或 RUNNABLE，则不执行任何操作，返回 false
    if (p->state == RUNNING || p->state == RUNNABLE || p->state == ZOMBIE ||
        (p->state == DEEPSLEEPING && onalert)) {
        release_sched_lock();
        return false;
    }

    // 如果进程的状态为 SLEEPING 或 UNUSED，将其状态设置为 RUNNABLE，
    // 重置等待时间，将其加入调度队列并返回 true
    if (p->state == SLEEPING || p->state == UNUSED ||
        (p->state == DEEPSLEEPING && !onalert)) {
        p->state = RUNNABLE;
        p->wait_time = 0;
        _insert_into_list(&rq[cid][p->true_priority + 4],
                          &p->schinfo.rq); // 加入调度队列
        release_sched_lock();
        return true;
    }

    // 如果状态不在以上范围内，触发 PANIC
    release_sched_lock();
    PANIC();

    return false; // 作为防御性返回
}

static void update_this_state(enum procstate new_state)
{
    thisproc()->currfuncid = 5;
    // TODO: if you use template sched function, you should implement this routinue
    // update the state of current process to new_state, and modify the sched queue if necessary
    // 更新当前进程的状态

    Proc *p = thisproc();
    u64 c_id = cpuid();
    // 如果新状态是SLEEPING或者ZOMBIE，需要将其从调度队列中移除
    if (p != cpus[c_id].sched.idle &&
        (p->state == RUNNING || p->state == RUNNABLE)) {
        _detach_from_list(&p->schinfo.rq);
    }
    p->state = new_state;
    if (thisproc() != cpus[c_id].sched.idle &&
        (new_state == RUNNING || new_state == RUNNABLE)) {
        thisproc()->wait_time = 0;
        _insert_into_list(&rq[c_id][p->true_priority + 4], &p->schinfo.rq);
    }
    if (!sched_timer[c_id].triggered) {
        cancel_cpu_timer(&sched_timer[c_id]);
    }
    set_cpu_timer(&sched_timer[c_id]);
}

// static Proc *pick_next()
// {
//     Proc *chosen_proc = NULL;
//     int max_wait_time = -1;

//     _for_in_list(p, &rq[cpuid()]) {
//         auto proc = container_of(p, struct Proc, schinfo.rq[cpuid()]);

//         // 更新等待时间并选择等待时间最长的进程
//         if (proc->state == RUNNABLE) {
//             proc->wait_time++;

//             if (proc->wait_time > max_wait_time) {
//                 max_wait_time = proc->wait_time;
//                 chosen_proc = proc;
//             }
//         }
//     }

//     // 如果没有找到合适的进程，返回 idle 进程
//     return chosen_proc ? chosen_proc : cpus[cpuid()].sched.idle;
// }

static Proc *pick_next()
{
    thisproc()->currfuncid = 6;
    //选择优先级最高(RMS：周期最短)且未完成的
    Proc *previous_proc = thisproc();
    Proc *chosen_proc = NULL;
    int max_priority = -5;
    if (!previous_proc->idle) {
        Insched_timer_handler(&sched_timer[cpuid()]);
    }

    for (int i = NUM_PRIORITY - 1; i >= 0; i--) {
        if (&rq[cpuid()][i] == rq[cpuid()][i].next)
            continue;
        auto p = rq[cpuid()][i].next;
        while (p != &rq[cpuid()][i]) {
            auto proc = container_of(p, struct Proc, schinfo.rq);
            // 更新等待时间并选择等待时间最长的进程
            if (proc->state == RUNNABLE) {
                //printk("idle:%d,period:%d,completed:%d\n",proc->idle,proc->period,proc->complete);
                // if(proc->complete){
                //     printk("Task:%d completed\n",proc->pid);
                // }
                // if (!proc->idle&&proc->period < min_period && !proc->complete) {
                //     //  if(proc->pid==1){
                //     //      printk("pid:%d,idle:%d,period:%d,completed:%d\n",proc->pid,proc->idle,proc->period,proc->complete);
                //     //  }

                //     min_period = proc->period;
                //     chosen_proc = proc;
                // }
                if ((!proc->idle && proc->true_priority > max_priority &&
                     !proc->complete) ||
                    previous_proc->idle) {
                    //  if(proc->pid==1){
                    //      printk("pid:%d,idle:%d,period:%d,completed:%d\n",proc->pid,proc->idle,proc->period,proc->complete);
                    //  }

                    max_priority = proc->true_priority;
                    chosen_proc = proc;
                }
            }
            p = p->next;
        }
        if (chosen_proc != NULL)
            break;
    }
    if (!previous_proc->idle && previous_proc && chosen_proc &&
        !previous_proc->complete && previous_proc != chosen_proc) {
        printk("Info: Process preemption occurred at CPU %lld.\n", cpuid());
        if (previous_proc) {
            printk("Previous Process Info:\n");
            printk("PID: %d, Exec Time: %d, Current Time: %d, Period: %d, Execed Time: %d,priority:%d\n",
                   previous_proc->pid, previous_proc->exec_time,
                   previous_proc->current_time, previous_proc->period,
                   previous_proc->execed_time, previous_proc->prim_priority);

            // 打印选中的进程的信息
            printk("Chosen Process Info:\n");
            printk("PID: %d, Exec Time: %d, Current Time: %d, Period: %d, Execed Time: %d,priority:%d\n",
                   chosen_proc->pid, chosen_proc->exec_time,
                   chosen_proc->current_time, chosen_proc->period,
                   chosen_proc->execed_time, chosen_proc->prim_priority);
        }
        printk("\n\n");
    }

    // if (previous_proc&&previous_proc->period) {
    //     printk("Previous Process Info:\n");
    //     printk("PID: %d, Exec Time: %d, Current Time: %d, Period: %d, Execed Time: %d\n",
    //            previous_proc->pid, previous_proc->exec_time, previous_proc->current_time,
    //            previous_proc->period, previous_proc->execed_time);
    // }

    // // 打印当前被选中的进程的信息
    // if (chosen_proc&&chosen_proc->period) {
    //     printk("Chosen Process Info:\n");
    //     printk("PID: %d, Exec Time: %d, Current Time: %d, Period: %d, Execed Time: %d\n",
    //            chosen_proc->pid, chosen_proc->exec_time, chosen_proc->current_time,
    //            chosen_proc->period, chosen_proc->execed_time);
    // }
    // 如果没有找到合适的进程，返回 idle 进程
    return chosen_proc ? chosen_proc : cpus[cpuid()].sched.idle;
}

static void update_this_proc(Proc *p)
{
    // TODO: you should implement this routinue
    // update thisproc to the choosen process
    //reset_clock(1000);//重置时钟中断，保证享有一个完整的时间片
    cpus[cpuid()].sched.thisproc = p;
}

// A simple scheduler.
// You are allowed to replace it with whatever you like.
// call with sched_lock
void sched(enum procstate new_state)
{
    thisproc()->currfuncid = 4;
    //TODO:修改你的sched代码，使其能够保证，如果当前进程带有killed标记，且new state不为zombie，则调度器直接返回，不做任何操作。
    auto this = thisproc();
    ASSERT(this->state == RUNNING);
    // 检查当前进程是否被标记为 killed 且 new_state 不是 ZOMBIE
    if (this->killed && new_state != ZOMBIE) {
        // 如果当前进程被杀死且新状态不是 ZOMBIE，直接返回
        release_sched_lock();
        return; // 直接返回，不做任何操作
    }
    update_this_state(new_state);
    auto next = pick_next();
    update_this_proc(next);
    ASSERT(next->state == RUNNABLE);
    next->state = RUNNING;
    if (next != this) {
        attach_pgdir(&next->pgdir);
        swtch(next->kcontext, &this->kcontext);
    }
    release_sched_lock();
}

u64 proc_entry(void (*entry)(u64), u64 arg)
{
    release_sched_lock();
    set_return_addr(entry);
    return arg;
}

void sleep(int delay)
{
    volatile int counter = 0;
    for (int i = 0; i < delay * 100000; i++) {
        counter++;
    }
}

// int fork_for_RT(int cid){
//     Proc * new_proc = fork();
//     if(new_proc == NULL)
//     {
//         return -1;
//     }
//     acquire_spinlock(&new_proc->plock);
//     _detach_from_list(&new_proc->schinfo.rq);
//     release_spinlock(&new_proc->plock);
//     acquire_sched_lock();
//     _insert_into_list(&rq[cid],&new_proc->schinfo.rq);
//     release_sched_lock();
//     return 0;
// }