#include <kernel/syscall.h>
#include <kernel/sched.h>
#include <kernel/printk.h>
#include <common/sem.h>
#include <test/test.h>
#include <aarch64/intrinsic.h>

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Woverride-init"

void *syscall_table[NR_SYSCALL] = {
    [0 ... NR_SYSCALL - 1] = NULL,
    [SYS_myreport] = (void *)syscall_myreport,
};

void syscall_entry(UserContext *context)
{
    // TODO
    // Invoke syscall_table[id] with args and set the return value.
    // id is stored in x8. args are stored in x0-x5. return value is stored in x0.
    // be sure to check the range of id. if id >= NR_SYSCALL, panic.
    u64 id = context->x[8];  // 从 x8 获取系统调用 ID
    if (id >= NR_SYSCALL) {
        PANIC();  // 如果 ID 超出范围，则触发 Panic
    }

    // 提取参数，分别存储在 x0 到 x5 中
    u64 arg0 = context->x[0];
    u64 arg1 = context->x[1];
    u64 arg2 = context->x[2];
    u64 arg3 = context->x[3];
    u64 arg4 = context->x[4];
    u64 arg5 = context->x[5];

    // 调用相应的系统调用
    if (syscall_table[id] != NULL) {
        // 调用系统调用并将返回值保存到 x0
        context->x[0] = ((u64 (*)(u64, u64, u64, u64, u64,u64))syscall_table[id])(arg0, arg1, arg2, arg3, arg4, arg5);
    } else {
        PANIC();  // 如果系统调用函数不存在，则触发 Panic
    }
}

#pragma GCC diagnostic pop